/* Copyright (C) 2003-2004 Datapark corp. All rights reserved.
   Copyright (C) 2000-2002 Lavtech.com corp. All rights reserved.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA 
*/

#include "dps_config.h"

#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include "dps_uniconv.h"
#include "dps_sgml.h"

/* UTF8 RFC 2279 */

int dps_mb_wc_utf8 (DPS_CONV *conv, DPS_CHARSET *cs, dpsunicode_t *pwc, const unsigned char *s, const unsigned char *end) {
  unsigned char c = s[0];
  int n = end - s;
  const unsigned char *p;
  unsigned char *e, z;
  unsigned int sw;
  
  conv->icodes = conv->ocodes = 1;

  if (c < 0x80) {
    if ( (*s == '&' && ((conv->flags & DPS_RECODE_HTML)||(conv->flags & DPS_RECODE_URL)) ) ||
	 (*s == '!' && (conv->flags & DPS_RECODE_URL)) ) {
      /*if ((p = strchr(s, ';')) != NULL)*/ {
	if (s[1] == '#') {
	  p = s + 2;
          if (s[2] == 'x' || s[2] == 'X') sscanf(s + 3, "%x", &sw);
	  else  sscanf(s + 2, "%d", &sw);
	  *pwc = (dpsunicode_t)sw;
	} else {
	  p = s + 1;
	  for(e = s + 1 ; (e - s < DPS_MAX_SGML_LEN) && (((*e<='z')&&(*e>='a'))||((*e<='Z')&&(*e>='A'))); e++);
	  if (!(conv->flags & DPS_RECODE_URL) || (*e == ';')) {
	    z = *e;
	    *e = '\0';
	    *pwc = DpsSgmlToUni(s + 1);
	    *e = z;
	  } else *pwc = 0;
	}
	if (*pwc) {
	  for (; isalpha(*p) || isdigit(*p); p++);
	  if (*p == ';') p++;
	  return conv->icodes = (p - s + 1);
	}
      }
    }
    *pwc = c;
    return 1;
  } else if (c < 0xc2) {
    return DPS_CHARSET_ILSEQ;
  } else if (c < 0xe0) {
    if (n < 2)return DPS_CHARSET_TOOFEW(0);
    if (!((s[1] ^ 0x80) < 0x40))return DPS_CHARSET_ILSEQ2;
    *pwc = ((dpsunicode_t) (c & 0x1f) << 6) | (dpsunicode_t) (s[1] ^ 0x80);
    return conv->icodes = 2;
  } else if (c < 0xf0) {
    if (n < 3)return DPS_CHARSET_TOOFEW(0);
    if (!((s[1] ^ 0x80) < 0x40 && (s[2] ^ 0x80) < 0x40 && (c >= 0xe1 || s[1] >= 0xa0)))
      return DPS_CHARSET_ILSEQ3;
    *pwc = ((dpsunicode_t) (c & 0x0f) << 12) | ((dpsunicode_t) (s[1] ^ 0x80) << 6) | (dpsunicode_t) (s[2] ^ 0x80);
    return conv->icodes = 3;
  } else if (c < 0xf8 && sizeof(dpsunicode_t)*8 >= 32) {
    if (n < 4)return DPS_CHARSET_TOOFEW(0);
    if (!((s[1] ^ 0x80) < 0x40 && (s[2] ^ 0x80) < 0x40 && (s[3] ^ 0x80) < 0x40 && (c >= 0xf1 || s[1] >= 0x90)))
      return DPS_CHARSET_ILSEQ4;
    *pwc = ((dpsunicode_t) (c & 0x07) << 18) | ((dpsunicode_t) (s[1] ^ 0x80) << 12) | ((dpsunicode_t) (s[2] ^ 0x80) << 6) | (dpsunicode_t) (s[3] ^ 0x80);
    return conv->icodes = 4;
  } else if (c < 0xfc && sizeof(dpsunicode_t)*8 >= 32) {
    if (n < 5)return DPS_CHARSET_TOOFEW(0);
    if (!((s[1] ^ 0x80) < 0x40 && (s[2] ^ 0x80) < 0x40 && (s[3] ^ 0x80) < 0x40 && (s[4] ^ 0x80) < 0x40 && (c >= 0xf9 || s[1] >= 0x88)))
      return DPS_CHARSET_ILSEQ5;
    *pwc = ((dpsunicode_t) (c & 0x03) << 24)
      | ((dpsunicode_t) (s[1] ^ 0x80) << 18)
      | ((dpsunicode_t) (s[2] ^ 0x80) << 12)
      | ((dpsunicode_t) (s[3] ^ 0x80) << 6)
      | (dpsunicode_t) (s[4] ^ 0x80);
    return conv->icodes = 5;
  } else if (c < 0xfe && sizeof(dpsunicode_t)*8 >= 32) {
     if (n < 6)return DPS_CHARSET_TOOFEW(0);
    if (!((s[1] ^ 0x80) < 0x40 && (s[2] ^ 0x80) < 0x40
      && (s[3] ^ 0x80) < 0x40 && (s[4] ^ 0x80) < 0x40
      && (s[5] ^ 0x80) < 0x40
      && (c >= 0xfd || s[1] >= 0x84)))
      return DPS_CHARSET_ILSEQ6;
    *pwc = ((dpsunicode_t) (c & 0x01) << 30)
      | ((dpsunicode_t) (s[1] ^ 0x80) << 24)
      | ((dpsunicode_t) (s[2] ^ 0x80) << 18)
      | ((dpsunicode_t) (s[3] ^ 0x80) << 12)
      | ((dpsunicode_t) (s[4] ^ 0x80) << 6)
      | (dpsunicode_t) (s[5] ^ 0x80);
    return conv->icodes = 6;
  } else
    return DPS_CHARSET_ILSEQ;
}



int dps_wc_mb_utf8(DPS_CONV *conv, DPS_CHARSET *cs, const dpsunicode_t *pwc, unsigned char *r, unsigned char *e) {
  int count;
  dpsunicode_t wc = *pwc;

  conv->icodes = conv->ocodes = 1;

  if (wc < 0x80) {
    r[0] = wc;
    if ((conv->flags & DPS_RECODE_HTML) && (r[0] == '"' || r[0] == '&' || r[0] == '<' || r[0] == '>')) 
      return DPS_CHARSET_ILUNI;
    if ((conv->flags & DPS_RECODE_URL) && (r[0] == '!')) 
      return DPS_CHARSET_ILUNI;
    return 1;
  }
  else if (wc < 0x800) 
    count = 2;
  else if (wc < 0x10000) 
    count = 3;
  else if (wc < 0x200000) 
    count = 4;
  else if (wc < 0x4000000) 
    count = 5;
  else if (wc <= 0x7fffffff) 
    count = 6;
  else 
    return DPS_CHARSET_ILUNI;
  
  if ( r+count > e)
    return DPS_CHARSET_TOOSMALL;
  
  switch (count) { /* Fall through all cases. */
    case 6: r[5] = 0x80 | (wc & 0x3f); wc = wc >> 6; wc |= 0x4000000;
    case 5: r[4] = 0x80 | (wc & 0x3f); wc = wc >> 6; wc |= 0x200000;
    case 4: r[3] = 0x80 | (wc & 0x3f); wc = wc >> 6; wc |= 0x10000;
    case 3: r[2] = 0x80 | (wc & 0x3f); wc = wc >> 6; wc |= 0x800;
    case 2: r[1] = 0x80 | (wc & 0x3f); wc = wc >> 6; wc |= 0xC0;
    case 1: r[0] = wc;
  }
  return conv->ocodes = count;
}

